# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import os
import signal
import sys
import time
from typing import TYPE_CHECKING, Optional

from rich.console import Console
from rich.status import Status
from rich.theme import Theme
from rich.traceback import install

import rucio.client.commands as commands
from rucio import version
from rucio.client.commands.bin_legacy.rucio import get_client, ping, test_server, whoami_account
from rucio.client.commands.command_base import CommandBase
from rucio.client.commands.utils import exception_handler, setup_gfal2_logger, signal_handler
from rucio.client.richclient import MAX_TRACEBACK_WIDTH, MIN_CONSOLE_WIDTH, CLITheme, get_cli_config, get_pager, setup_rich_logger
from rucio.common.extra import import_extras
from rucio.common.utils import setup_logger

if TYPE_CHECKING:
    from argparse import Namespace
    from logging import Logger

    from rucio.client.client import Client

EXTRA_MODULES = import_extras(["argcomplete"])

if EXTRA_MODULES["argcomplete"]:
    import argcomplete  # pylint: disable=E0401

SUCCESS, FAILURE = 0, 1


class Commands:
    def __init__(self, logger: "Logger", client: "Client", args: "Namespace", console: Console, spinner: Status) -> None:
        self.logger = logger
        self.args = args
        self.client = client
        self.console = console
        self.spinner = spinner

    @staticmethod
    def _all_commands() -> dict[str, type[commands.CommandBase]]:
        # Look for all the CommandBase'd child classes in this folder and add their parsers and child parsers
        custom_commands = ['LifetimeException', "TestServer"]  # These ones have custom names
        command_map = {child.__name__.lower(): child for child in commands.CommandBase.__subclasses__() if child.__name__ not in custom_commands}
        command_map["lifetime-exception"] = commands.LifetimeException
        command_map["test-server"] = TestServer
        return command_map

    @staticmethod
    def _add_parsers() -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(description='CLI Rucio Client. (Use --legacy to view CLI from <36.0)')
        Commands._main_parser(parser)

        subparsers = parser.add_subparsers(dest="command", help="Command to execute, see `{command} -h` for more details and subcommands.")
        for command in Commands._all_commands().values():
            command(None, None, None, None, None).parser(subparsers)  # type: ignore

        return parser

    @staticmethod
    def _main_parser(parser: argparse.ArgumentParser) -> None:
        main_args = parser.add_argument_group("Main Arguments")

        main_args.add_argument("--version", action="version", version="%(prog)s " + version.version_string())

        main_args.add_argument("--config", help="The Rucio configuration file to use.")
        main_args.add_argument("--verbose", "-v", default=False, action="store_true", help="Print more verbose output.")
        main_args.add_argument("-H", "--host", metavar="ADDRESS", help="The Rucio API host.")
        main_args.add_argument("--auth-host", metavar="ADDRESS", help="The Rucio Authentication host.")
        main_args.add_argument("-a", "--account", dest="issuer", help="Rucio account to use.")
        main_args.add_argument("-S", "--auth-strategy", help="Authentication strategy (userpass, x509...)")
        main_args.add_argument("-T", "--timeout", type=float, help="Set all timeout values to seconds.")
        main_args.add_argument("--user-agent", "-U", dest="user_agent", default="rucio-clients", help="Rucio User Agent")
        main_args.add_argument("--vo", metavar="VO", help="VO to authenticate at. Only used in multi-VO mode.")
        main_args.add_argument("--no-pager", dest="no_pager", default=False, action='store_true', help=argparse.SUPPRESS)

        auth_args = parser.add_argument_group("Authentication Settings")

        auth_args.add_argument("-u", "--user", dest="username", help="username")
        auth_args.add_argument("-pwd", "--password", help="password")
        # Options for defining remaining OIDC parameters
        auth_args.add_argument("--oidc-user", dest="oidc_username", help="OIDC username")
        auth_args.add_argument("--oidc-password", help="OIDC password")
        auth_args.add_argument(
            "--oidc-scope",
            default="openid profile",
            help="""
                Defines which (OIDC) information user will share with Rucio. Rucio requires at least -sc='openid profile'.
                To request refresh token for Rucio, scope must include 'openid offline_access' and there must be no active
                access token saved on the side of the currently used Rucio Client.
            """
        )
        auth_args.add_argument("--oidc-audience", help="Defines which audience are tokens requested for.")
        auth_args.add_argument(
            "--oidc-auto",
            default=False,
            action="store_true",
            help="""
                If not specified, username and password credentials are not required and users will be given a URL to use in their browser.
                If specified, the users explicitly trust Rucio with their IdP credentials.
             """
        )
        auth_args.add_argument(
            "--oidc-polling",
            default=False,
            action="store_true",
            help="""
                If not specified, user will be asked to enter a code returned by the browser to the command line.
                If --polling is set, Rucio Client should get the token without any further interaction of the user.
                This option is active only if --auto is *not* specified.
            """
        )
        auth_args.add_argument(
            "--oidc-refresh-lifetime",
            help="""
                Max lifetime in hours for this access token; the token will be refreshed by an asynchronous Rucio daemon.
                If not specified, refresh will be stopped after 4 days.
                This option is effective only if --oidc-scope includes offline_access scope for a refresh token to be granted to Rucio.
            """
        )
        auth_args.add_argument("--oidc-issuer", help="Defines which Identity Provider is going to be used. The issuer string must correspond to the keys configured in the /etc/idpsecrets.json auth server configuration file.")  # NOQA: W503

        # Options for the x509  auth_strategy
        auth_args.add_argument("--certificate", help="Client certificate file.")
        auth_args.add_argument('--client-key', dest='client_key', default=None, help='Client key for x509 Authentication.')
        auth_args.add_argument("--ca-certificate", help="CA certificate to verify peer against (SSL).")

    def _run_command(self) -> Optional[int]:
        try:
            command_class = Commands._all_commands()[self.args.command]
        except KeyError as e:
            if self.args.command is not None:
                self.logger.error(f"Cannot find command {self.args.command}: {e}")
            sys.exit(FAILURE)

        return exception_handler(command_class(self.client, self.args, self.logger, self.console, self.spinner))()

    def __call__(self) -> None:
        self.logger.debug("Running a command with the following arguments: %s" % vars(self.args))
        start_time = time.time()
        self._run_command()
        end_time = time.time()

        self.logger.debug("Completed in %-0.4f sec." % (end_time - start_time))


class Ping(CommandBase):
    def _operations(self):
        return {}

    def module_help(self) -> str:
        return ""

    def usage_example(self) -> list[str]:
        return []

    def parser(self, subparser):
        command_parser = subparser.add_parser("ping", description="Ping the server", formatter_class=argparse.RawDescriptionHelpFormatter)
        return command_parser

    def __call__(self):
        ping(self.args, self.client, self.logger, self.console, self.spinner)


class Whoami(CommandBase):
    def _operations(self):
        return {}

    def module_help(self) -> str:
        return ""

    def usage_example(self) -> list[str]:
        return []

    def parser(self, subparser):
        command_parser = subparser.add_parser("whoami", description="See login information, test credentials", formatter_class=argparse.RawDescriptionHelpFormatter)
        return command_parser

    def __call__(self):
        whoami_account(self.args, self.client, self.logger, self.console, self.spinner)


class TestServer(CommandBase):
    def _operations(self):
        return {}

    def module_help(self) -> str:
        return ""

    def usage_example(self) -> list[str]:
        return []

    def parser(self, subparser):
        command_parser = subparser.add_parser("test-server", description="Test client against the server", formatter_class=argparse.RawDescriptionHelpFormatter)
        return command_parser

    def __call__(self):
        test_server(self.args, self.client, self.logger, self.console, self.spinner)


def main():
    cli_config = get_cli_config()
    console = Console(theme=Theme(CLITheme.LOG_THEMES), soft_wrap=True)
    console.width = max(MIN_CONSOLE_WIDTH, console.width)

    spinner = Status('Initializing spinner', spinner=CLITheme.SPINNER, spinner_style=CLITheme.SPINNER_STYLE, console=console)
    pager = get_pager()

    parser_object = Commands._add_parsers()
    if EXTRA_MODULES["argcomplete"]:
        argcomplete.autocomplete(parser_object)
    args = parser_object.parse_args()
    if args.config is not None:
        os.environ["RUCIO_CONFIG"] = args.config
    if args.command is None:
        parser_object.print_help()
        sys.exit(FAILURE)

    setup_gfal2_logger()

    if cli_config == 'rich':
        install(console=console, word_wrap=True, width=min(console.width, MAX_TRACEBACK_WIDTH))  # Make rich exception tracebacks the default.
        logger = setup_rich_logger(module_name=__name__, logger_name='user', verbose=args.verbose, console=console)

    else:
        logger = setup_logger(module_name=__name__, logger_name="user", verbose=args.verbose)

    client = get_client(args, logger)
    signal.signal(signal.SIGINT, lambda sig, frame: signal_handler(sig, frame, logger))

    Commands(logger, client, args, console, spinner)()

    if cli_config == 'rich':
        spinner.stop()

    if console.is_terminal and not args.no_pager:
        command_output = console.end_capture()
        if command_output != '':
            signal.signal(signal.SIGINT, signal.SIG_IGN)  # Do not allow the user to stop the program while sending to pager
            pager(command_output)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from abc import ABC, abstractmethod
from argparse import ArgumentError, RawDescriptionHelpFormatter
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace, _SubParsersAction
    from logging import Logger

    from rich.console import Console
    from rich.status import Status

    from rucio.client.client import Client
    from rucio.client.commands.utils import OperationDict


class CommandBase(ABC):
    def __init__(self, client: "Client", args: "Namespace", logger: "Logger", console: "Console", spinner: "Status") -> None:
        self.COMMAND_NAME = sys.argv[0].split("/")[-1]
        self.PARSER_NAME = self.__class__.__name__.lower()
        self.client = client
        self.args = args
        self.logger = logger
        self.console = console
        self.spinner = spinner

    @abstractmethod
    def module_help(self) -> str:
        """
        Description of the command
        """
        raise NotImplementedError

    @abstractmethod
    def usage_example(self) -> list[str]:
        """
        List of examples. Examples should be in the form `$ rucio <command> ...  # A quick description of what this does`
        """
        raise NotImplementedError

    @abstractmethod
    def _operations(self) -> dict[str, "OperationDict"]:
        """
        Dictionary description of the different implemented operations with their associated call and documentation.
        Example:
            If you want to have: $ rucio do something --kwarg this

            class Do(CommandBase):
                ...

                def _operations(self):
                    return {"something": "call": self.something, "namespace": self.something_namespace, "docs": "Does Something"}

                def something(self):
                    ...

                def something_namespace(self, parser):
                    parser.add_argument("--kwarg", choices={"this", "that"})
        """
        raise NotImplementedError

    def implemented_subcommands(self) -> dict[str, type["CommandBase"]]:
        """
        Provide a mapping of subcommands. Leave empty if command has no subcommands.
        Example:
          Running commands: $ rucio foo bar --kwarg

          class Foo(CommandBase):
            ...
            def implemented_subcommands(self):
                return {"bar": Bar}

          class Bar(Foo):
            ...
        """
        return {}

    def _help(self) -> str:
        """
        Create a help string including the module help, operation options, subcommand options, a usage example, and stating the default operation.
        """
        help = f"{self.module_help()}\n"

        help += "\nMatching Commands:\n"
        for operation_name in self._operations().keys():
            help += f"rucio {self.PARSER_NAME} {operation_name}\n"

        for subcommand_name, subcommand in self.implemented_subcommands().items():
            for operation_name in subcommand(self.client, self.args, self.logger, self.console, self.spinner)._operations().keys():
                help += f"rucio {self.PARSER_NAME} {subcommand_name} {operation_name}\n"

        examples = "\n".join(self.usage_example())
        help += f"\n\nUsage Example:\n{examples}"

        return help

    def parser(self, subparser: "_SubParsersAction[ArgumentParser]") -> "ArgumentParser":
        """
        Create a parser for a given argument. Adds the positional arguments (operations and subcommands) and the namespaces for default operations
        """
        command_parser = subparser.add_parser(self.PARSER_NAME, description=self._help(), formatter_class=RawDescriptionHelpFormatter)
        subcommand_parser = command_parser.add_subparsers(dest=f"{self.PARSER_NAME}_subcommand")

        try:
            subcommands = self.implemented_subcommands()
            for _, subcommand in subcommands.items():
                # Subparsers are given their own call to avoid recursionErrors.
                subcommand(self.client, self.args, self.logger, self.console, self.spinner).subparser(subcommand_parser)
        # If the command has no subcommands
        except (NotImplementedError, TypeError):
            pass

        operations = self._operations()
        for name, operation in operations.items():
            help = operation.get("docs", "")
            parser = subcommand_parser.add_parser(name, help=help)
            namespace = operation.get("namespace")
            if namespace is not None:
                namespace(parser)

            # Just get the namespace argument if there isn't a specific namespace for a operation
            elif hasattr(self, "namespace"):
                try:
                    self.namespace(command_parser)  # type: ignore
                except ArgumentError:  # Namespace has already been added
                    pass

        return command_parser

    def subparser(self, parser: "_SubParsersAction[ArgumentParser]"):
        """
        Create a parser for a new subcommand. Add the default operation namespace to the top level menu.
        """
        subparser = parser.add_parser(self.PARSER_NAME, description=self._help(), help=self.module_help(), formatter_class=RawDescriptionHelpFormatter)

        operations = self._operations()
        subsubparser = subparser.add_subparsers(dest=f"{self.PARSER_NAME}_subcommand")

        for name, operation in operations.items():
            help = operation.get("docs", "")
            operation_parser = subsubparser.add_parser(name, help=help)
            namespace = operation.get("namespace")
            if namespace is not None:
                namespace(operation_parser)
            elif hasattr(self, "namespace"):
                self.namespace(operation_parser)  # type: ignore

        return subparser

    def _execute_subcommand(self, subcommands: dict, requested_subcommand: str, requested_verb: Optional[str] = None):
        if requested_subcommand in subcommands.keys():
            subcommand_client = subcommands[requested_subcommand](self.client, self.args, self.logger, self.console, self.spinner)

            subcommand_operations = subcommand_client._operations()
            if requested_verb in subcommand_operations.keys():
                subcommand_operations[requested_verb]["call"]()
            elif requested_verb is not None:
                raise NotImplementedError

    def __call__(self):
        subcommands = self.implemented_subcommands()
        operations = self._operations()

        requested_subcommand = vars(self.args).get(f"{self.PARSER_NAME}_subcommand")
        requested_verb = vars(self.args).get("verb")

        # If the subcommand has multiple positional args, they can be contained in the {name}_subcommand dict
        if (requested_verb is None) and (requested_subcommand is not None):
            requested_verb = vars(self.args).get(f"{requested_subcommand}_subcommand")

        if (requested_subcommand is not None) and (subcommands is not None):
            if requested_subcommand in operations.keys():
                # Treat the subcommand as the operational verb
                operations[requested_subcommand]["call"]()

            elif subcommands is not None:
                self._execute_subcommand(subcommands, requested_subcommand, requested_verb)

            else:  # Should not be possible, but better to catch the error than not.
                raise NotImplementedError



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from argparse import SUPPRESS
from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import add_rule, delete_rule, info_rule, list_rules, list_rules_history, move_rule, update_rule
from rucio.client.commands.command_base import CommandBase

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class Rule(CommandBase):
    def module_help(self) -> str:
        return "Create rules that require a number of replicas at a defined set of remote Rucio Storage Elements (RSEs). Rules will initialize transfers between sites"

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "add": {"call": self.add, "docs": "Add replication rule to define number of replicas at sites", "namespace": self.add_namespace},
            "remove": {"call": self.remove, "docs": "Delete a replication rule", "namespace": self.remove_namespace},
            "show": {"call": self.show, "docs": "Retrieve information about a specific rule", "namespace": self.show_namespace},
            "history": {"call": self.history, "docs": "List history of different replica rules impacting a DID", "namespace": self._common_namespace},
            "update": {"call": self.update, "docs": "Update replication rule, can be used to move a rule from one RSE to another", "namespace": self.update_namespace},
            "list": {"call": self.list_, "docs": "List replication rules", "namespace": self.list_namespace},
        }

    def usage_example(self) -> list[str]:
        return [
            "$ rucio rule add -d user.jdoe:test_did --copies 2 --rses SPAINSITES  # Create a rule that requires two copies of a did limited to Spanish Sites",
            "$ rucio rule list --did user.jdoe:test_did  # show rules impacting a DID",
            "$ rucio rule show --rule-id rule123456  # View a detailed overview for an existing rule",
            "$ rucio rule remove --rule-id rule123456  # Deactivate a rule",
            "$ rucio rule update --rule-id rule123456 --suspend  # Suspend the execution of a rule",
            "$ rucio rule update --rule-id rule123456 --move --rses NewRSE # Copy an existing rule to a new RSE",
        ]

    def _common_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-a", "--account", dest="rule_account", action="store", help="The account of the rule")
        parser.add_argument("--activity", action="store", help="Activity to be used (e.g. User, Data Consolidation)")  # TODO More info on this
        parser.add_argument("--rule-id", help="The rule ID, for accessing an existing rule.")
        parser.add_argument("--lifetime", dest="lifetime", action="store", type=int, help="Rule lifetime (in seconds)")
        parser.add_argument("--locked", action="store_true", help="Set the rule to locked")  # TODO More info on this
        parser.add_argument("--source-replica-expression", help="RSE Expression for RSEs to be considered for source replicas")
        parser.add_argument("--comment", dest="comment", action="store", help="Comment about the replication rule")

    def add_namespace(self, parser: "ArgumentParser") -> None:
        self._common_namespace(parser)
        parser.add_argument("--rses", "--rse-exp", dest="rse_expression", action="store", help="The RSE expression. Must be specified if a DID is provided")
        parser.add_argument("-d", "--did", dest="dids", nargs="+", help="DID(s) to apply the rule to")
        parser.add_argument("--copies", type=int, help="Number of copies")
        parser.add_argument("--weight", help="RSE Weight")  # TODO What does this do
        parser.add_argument("--grouping", choices=["DATASET", "ALL", "NONE"], help="Rule grouping")  # TODO What does this do
        parser.add_argument("--notify", action="store", help="Notification strategy : Y (Yes), N (No), C (Close)", choices={"Y", "N", "C"})
        parser.add_argument("--ask-approval", action="store_true", help="Ask for rule approval")
        parser.add_argument("--asynchronous", action="store_true", help="Create rule asynchronously")
        parser.add_argument("--delay-injection", type=int, help="Delay (in seconds) to wait before starting applying the rule. This option implies --asynchronous")
        parser.add_argument("--skip-duplicates", action="store_true", help="Skip duplicate rules")

    def remove_namespace(self, parser: "ArgumentParser") -> None:
        self._common_namespace(parser)
        parser.add_argument("--purge-replicas", action="store_true", help="Purge rule replicas")
        parser.add_argument("--all", dest="delete_all", action="store_true", default=False, help="Delete all the rules, even the ones that are not owned by the account")

    def show_namespace(self, parser: "ArgumentParser") -> None:
        self._common_namespace(parser)
        parser.add_argument("--examine", action="store_true", help="Detailed analysis of transfer errors")

    def list_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-a", "--account", dest="rule_account", action="store", help="The account of the rule")
        parser.add_argument("-d", "--did", help="DIDs to look for rules")
        parser.add_argument("--traverse", action="store_true", help="Traverse the did tree and search for rules affecting this did")
        parser.add_argument("--csv", action="store_true", help="Write output to a CSV")
        parser.add_argument("--file", help="List associated rules of an affected file")
        parser.add_argument("--subscription", help="List by account and subscription name", metavar=("ACCOUNT", "SUBSCRIPTION"), nargs=2)
        parser.add_argument("--human", default=True, help=SUPPRESS)
        parser.add_argument("--rule-id", help=SUPPRESS)

    def update_namespace(self, parser: "ArgumentParser") -> None:
        self._common_namespace(parser)
        parser.add_argument("--rses", "--rse-exp", dest="rse_expression", help="RSE to change for the rule")
        parser.add_argument("--stuck", dest="state_stuck", action="store_true", help="Set state to STUCK")
        parser.add_argument("--suspend", dest="state_suspended", action="store_true", help="Set state to SUSPENDED")
        parser.add_argument("--cancel-requests", action="store_true", help="Cancel requests when setting rules to stuck")
        parser.add_argument("--priority", help="Priority of the requests of the rule")
        parser.add_argument("--child-rule-id", help='Child rule id of the rule. Use "None" to remove an existing parent/child relationship')
        parser.add_argument("--boost-rule", action="store_true", help="Quickens the transition of a rule from STUCK to REPLICATING")
        parser.add_argument("--move", action="store_true", help="Duplicate the existing replication rule to a different RSE")

    def list_(self):
        list_rules(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        add_rule(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_rule(self.args, self.client, self.logger, self.console, self.spinner)

    def show(self):
        info_rule(self.args, self.client, self.logger, self.console, self.spinner)

    def history(self):
        list_rules_history(self.args, self.client, self.logger, self.console, self.spinner)

    def update(self):
        self.args.rule_activity = self.args.activity
        if self.args.move:
            move_rule(self.args, self.client, self.logger, self.console, self.spinner)

        else:
            update_rule(self.args, self.client, self.logger, self.console, self.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio_admin import add_subscription, list_subscriptions, reevaluate_did_for_subscription, update_subscription
from rucio.client.commands.command_base import CommandBase

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class Subscription(CommandBase):
    def module_help(self) -> str:
        return "Automate processing of specific rules, will create new rules on a schedule"

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "List all active subscriptions", "namespace": self.list_namespace},
            "update": {"call": self.update, "docs": "Update an existing subscription", "namespace": self._common_namespace},
            "add": {"call": self.add, "docs": "Create a new subscription", "namespace": self._common_namespace},
            "touch": {"call": self.touch, "docs": "Reevaluate did for subscription", "namespace": self.touch_namespace},
        }

    def usage_example(self) -> list[str]:
        return [
            "$ rucio subscription add --lifetime 2 --account jdoe --priority 1 --name jdoes_txt_files_on_datadisk  # Create a new subscription to create new rules",
            "$ rucio subscription list --account jdoe # List subscriptions for jdoe's account. Shows rules created by this subscription",
        ]

    def _common_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--name", help="Subscription name, used to identify the subscription in the place of an ID")
        parser.add_argument("--filter", help='DID filter (eg \'{"scope": ["tests"], "project": ["data12_8TeV"]}\')')
        parser.add_argument("--rules", dest="replication_rules", help='Replication rules (eg \'[{"copies": 2, "rse_expression": "tier=2", "lifetime": 3600, "activity": "Functional Tests", "weight": "mou"}]\')')
        parser.add_argument("--comments", help="Comments on subscription")
        parser.add_argument("--lifetime", type=int, help="Subscription lifetime (in days)")
        parser.add_argument("-a", "--account", dest="subs_account", help="Account name")
        parser.add_argument("--priority", help="The priority of the subscription")

    def list_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--name", help="Subscription name, used to identify the subscription in the place of an ID")
        parser.add_argument("-a", "--account", dest="subs_account", help="Account name")
        parser.add_argument("--filter", help='DID filter (eg \'{"scope": ["tests"], "project": ["data12_8TeV"]}\')')
        parser.add_argument("--long", action="store_true", help="Show extra subscription information, including creation and expiration dates")

    def touch_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-d", "--did", dest="dids", help="List of DIDs (coma separated)", required=True)

    def list_(self):
        list_subscriptions(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        add_subscription(self.args, self.client, self.logger, self.console, self.spinner)

    def update(self):
        update_subscription(self.args, self.client, self.logger, self.console, self.spinner)

    def touch(self):
        reevaluate_did_for_subscription(self.args, self.client, self.logger, self.console, self.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import list_rses
from rucio.client.commands.bin_legacy.rucio_admin import (
    add_distance_rses,
    add_protocol_rse,
    add_qos_policy,
    add_rse,
    del_protocol_rse,
    delete_attribute_rse,
    delete_distance_rses,
    delete_limit_rse,
    delete_qos_policy,
    disable_rse,
    get_attribute_rse,
    get_distance_rses,
    info_rse,
    list_qos_policies,
    set_attribute_rse,
    set_limit_rse,
    update_distance_rses,
    update_rse,
)
from rucio.client.commands.command_base import CommandBase

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class RSE(CommandBase):
    def module_help(self) -> str:
        return "Manage Rucio Storage Elements - the sites where Rucio can place and access data."

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "Show all RSEs", "namespace": self.list_namespace},
            "show": {"call": self.show, "docs": "Show history of RSE usage (in terms of bytes and files)", "namespace": self.namespace},
            "remove": {"call": self.remove, "docs": "Disable an RSE", "namespace": self.namespace},
            "add": {"call": self.add, "docs": "Create a new RSE", "namespace": self.namespace},
            "update": {"call": self.update, "docs": "Update an existing RSE", "namespace": self.namespace},
        }

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {"attribute": Attribute, "distance": Distance, "protocol": Protocol, "limit": Limit, "qos": QOS}

    def usage_example(self) -> list[str]:
        return [
            "$ rucio rse list # Show all current RSEs, can also access with",
            "$ rucio rse list --rses 'deterministic=True' # Show all RSEs that match the RSE Expression 'deterministic=True'",
            "$ rucio rse remove --rse RemoveThisRSE  # Disable an RSE by name",
            "$ rucio rse add --rse CreateANewRSE  # add a new RSE named CreateANewRSE",
            "$ rucio rse update --rse rse123456 --setting deterministic --value False  # Make an RSE Non-Deterministic",
            "$ rucio rse show --rse rse123456  # See all the settings for rse123456"
        ]

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--rse", "--rse-name", help="RSE name", required=True)
        parser.add_argument("--non-deterministic", action="store_true", help="Create RSE in non-deterministic mode")

        parser.add_argument(
            "--setting",
            dest="param",
            help="RSE setting",
            choices={"deterministic", "rse_type", "staging_area", "volatile", "qos_class", "availability_delete", "availability_read", "availability_write", "city", "country_name", "latitude", "longitude", "region_code", "time_zone"},
        )
        parser.add_argument("--value", dest="value", help='Value for the new setting configuration. Use "", None or null to wipe the value')

    def list_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--rses", "--rse-exp", dest="rses", help="RSE name or expression")

    def list_(self):
        list_rses(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        add_rse(self.args, self.client, self.logger, self.console, self.spinner)

    def update(self):
        update_rse(self.args, self.client, self.logger, self.console, self.spinner)

    def show(self):
        info_rse(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        disable_rse(self.args, self.client, self.logger, self.console, self.spinner)


class Attribute(RSE):
    def module_help(self) -> str:
        return "Manage RSE Attributes as key/value pairs. \nCAUTION: the existing attributes can be overwritten"

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "Show existing attributes for an RSE"},
            "add": {"call": self.add, "docs": "Update an RSE's setting. Will overwrite existing settings"},
            "remove": {"call": self.remove, "docs": "Wipe an RSE's setting"}
        }

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--rse", "--rse-name", help="RSE name", required=True)
        parser.add_argument("--key", help="Attribute key")
        parser.add_argument("--value", help="Attribute value")

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def usage_example(self) -> list[str]:
        return [
            "$ rucio rse attribute list --rse ThisRSE  # Show all the attributes for a given RSE",
            "$ rucio rse attribute list --rse ThisRSE --key name # Show all the attribute 'name' for a given RSE",
            "$ rucio rse attribute add --rse ThisRSE --key given-attribute --value updated # Set the attribute 'given-attribute' to 'updated' for an RSE",
        ]

    def add(self):
        set_attribute_rse(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_attribute_rse(self.args, self.client, self.logger, self.console, self.spinner)

    def list_(self):
        get_attribute_rse(self.args, self.client, self.logger, self.console, self.spinner)


class Distance(RSE):
    def module_help(self) -> str:
        return "Manage distances between RSEs. Used for determining efficiency of transfers from RSE to RSE via multihop operations"

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "Show distances between RSEs"},
            "add": {"call": self.add, "docs": "Add or update a distance between RSEs"},
            "remove": {"call": self.remove, "docs": "Delete the distance between a pair of RSEs. The mandatory parameters are source and destination"},
            "set": {"call": self.set_, "docs": "Update the distance between a pair of RSE that already have a distance between them"},
        }

    def usage_example(self) -> list[str]:
        return [
            "$ rucio rse distance list --source rse1 --destination rse2  # View the existing distance between rse1 and rse2",
            "$ rucio rse distance remove --source rse1 --destination rse2  # Remove an existing distance",
            "$ rucio rse distance add --source rse1 --destination rse2 --distance 10  # Add the distance between two rses that do not already have a distance",
            "$ rucio rse distance set --source rse1 --destination rse2 --distance 20  # Update an existing distance",
        ]

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--source", dest="source", action="store", help="Source RSE name")
        parser.add_argument("--destination", dest="destination", action="store", help="Destination RSE name")
        parser.add_argument("--distance", dest="distance", type=int, help="Distance between RSEs")

    def list_(self):
        get_distance_rses(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        add_distance_rses(self.args, self.client, self.logger, self.console, self.spinner)

    def set_(self):
        update_distance_rses(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_distance_rses(self.args, self.client, self.logger, self.console, self.spinner)


class Protocol(RSE):
    def module_help(self) -> str:
        return "Manage RSE transfer and storage protocols. Use `$ rucio rse show` to view an RSE's existing protocols"

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def _operations(self) -> dict[str, "OperationDict"]:
        return {"add": {"call": self.add, "docs": "Create a new RSE transfer protocol"}, "remove": {"call": self.remove, "docs": "Remove an existing RSE protocol"}}

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--rse", "--rse-name", dest='rse', help="RSE name")
        parser.add_argument("--host", dest="hostname", help="Endpoint hostname")
        parser.add_argument("--scheme", help="Endpoint URL scheme")
        parser.add_argument("--prefix", help="Endpoint URL path prefix")
        parser.add_argument("--space-token", help="Space token name (SRM-only)")
        parser.add_argument("--web-service-path", help="Web service URL (SRM-only)")
        parser.add_argument("--port", type=int, help="URL port")
        parser.add_argument("--impl", default="rucio.rse.protocols.gfal.Default", help="Transfer protocol implementation to use")
        parser.add_argument("--domain-json", type=json.loads, help="JSON describing the WAN / LAN setup")
        parser.add_argument("--extended-attributes-json", dest="ext_attr_json", type=json.loads, help="JSON describing any extended attributes")

    def usage_example(self) -> list[str]:
        return [
            "$ rucio rse protocol --hostname jdoes.test.org --scheme gsiftp --prefix '/atlasdatadisk/rucio/' --port 8443 --rse JDOE_DATADISK  # Add a new protocol on jdoe.test.org that uses gsiftp",
            "$ rucio rse protocol remove --scheme gsiftp --rse JDOE_DATADISK # Remove the existing gsiftp protocol",
        ]

    def add(self):
        add_protocol_rse(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        del_protocol_rse(self.args, self.client, self.logger, self.console, self.spinner)


class Limit(RSE):
    def module_help(self) -> str:
        return "Manage storage size limits. Existing limits can be found with `$ rucio rse info`"

    def _operations(self) -> dict[str, "OperationDict"]:
        return {"add": {"call": self.add, "docs": "Add a storage limit"}, "remove": {"call": self.remove, "docs": "Remove an existing storage limit"}}

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--rse", "--rse-name", dest='rse', help="RSE name")
        parser.add_argument("--name", help="Name of the limit")
        parser.add_argument("--limit", dest="value", help="Value of the limit in bytes")

    def usage_example(self) -> list[str]:
        return ["$ rucio rse limit add --rse XRD1 --name MinFreeSpace --value 10000", "$ rucio rse limit --rse XRD3 --name MinFreeSpace"]

    def add(self):
        set_limit_rse(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_limit_rse(self.args, self.client, self.logger, self.console, self.spinner)


class QOS(RSE):
    def module_help(self) -> str:
        return "Interact with an RSE's QoS policy."

    def _operations(self) -> dict[str, "OperationDict"]:
        return {"list": {"call": self.list_, "docs": "Show existing QoS Policies"}, "add": {"call": self.add, "docs": "Add a new policy"}, "remove": {"call": self.remove, "docs": "Remove Policy"}}

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--rse", "--rse-name", dest="rse", help="RSE name")
        parser.add_argument("--policy", dest="qos_policy", help="QoS policy")

    def usage_example(self) -> list[str]:
        return [
            "$ rucio rse qospolicy list --rse JDOE_DATADISK  # List QoS Policy for a given RSE",
            "$ rucio rse qospolicy add --rse JDOE_DATADISK --policy SLOW_BUT_CHEAP  # Add a SLOW_BUT_CHEAP policy to the JDOE_DATADISK RSE",
            "$ rucio rse qospolicy remove --rse  JDOE_DATADISK --policy SLOW_BUT_CHEAP  # Remove the same policy",
        ]

    def add(self):
        add_qos_policy(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_qos_policy(self.args, self.client, self.logger, self.console, self.spinner)

    def list_(self):
        list_qos_policies(self.args, self.client, self.logger, self.console, self.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import os
from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import download
from rucio.client.commands.command_base import CommandBase
from rucio.common.config import config_get_float
from rucio.common.utils import StoreAndDeprecateWarningAction

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class Download(CommandBase):
    def module_help(self) -> str:
        return "Download files"

    def parser(self, parser: "argparse._SubParsersAction[ArgumentParser]") -> None:

        command_parser = parser.add_parser(self.PARSER_NAME, description=self._help(), formatter_class=argparse.RawDescriptionHelpFormatter)

        command_parser.add_argument("-d", "--did", nargs="*", dest="dids", help="DIDs to download, as a space separated list", required=True)
        command_parser.add_argument("--rses", "--rse-exp", help="The Rucio Storage Element (RSE) name or expression to download from", dest="rses")
        command_parser.add_argument("--dir", dest="dir", default=".", action="store", help="The directory to store the downloaded file")
        command_parser.add_argument("--allow-tape", action="store_true", default=False, help="Also consider tape endpoints as source of the download")
        command_parser.add_argument("--impl", dest="impl", action="store", help="Transfer protocol implementation to use (e.g: xrootd, gfal.NoRename, webdav, ssh.Rsync, rclone)")
        command_parser.add_argument("--protocol", action="store", help="Force the protocol to use.")
        command_parser.add_argument("--nrandom", type=int, action="store", help="Download N random files from the DID.")
        command_parser.add_argument("--ndownloader", type=int, default=3, action="store", help="Choose the number of parallel processes for download")
        command_parser.add_argument("--no-subdir", action="store_true", default=False, help="Don't create a subdirectory for the scope of the files")
        command_parser.add_argument("--pfn", dest="pfn", action="store", help="Specify the exact PFN for the download")
        command_parser.add_argument("--archive-did", action="store", dest="archive_did", help="Download from archive is transparent. This option is obsolete")
        command_parser.add_argument("--no-resolve-archives", action="store_true", default=False, help="If set archives will not be considered for download")
        command_parser.add_argument("--ignore-checksum", action="store_true", default=False, help="Don't validate checksum for downloaded files")
        command_parser.add_argument("--check-local-with-filesize-only", action="store_true", default=False, help="Don't use checksum verification for already downloaded files, use filesize instead")
        command_parser.add_argument(
            "--transfer-timeout",
            dest="transfer_timeout",
            type=float,
            action="store",
            default=config_get_float("download", "transfer_timeout", False, None),
            help="Transfer timeout (in seconds). Default: computed dynamically from --transfer-speed-timeout. If set to any value >= 0, --transfer-speed-timeout is ignored",
        )  # NOQA: E501
        command_parser.add_argument(
            "--transfer-speed-timeout",
            dest="transfer_speed_timeout",
            type=float,
            action="store",
            default=None,
            help="Minimum allowed average transfer speed (in KBps). Default: 500. Used to dynamically compute the timeout if --transfer-timeout not set. Is not supported for --pfn",
        )  # NOQA: E501
        command_parser.add_argument("--aria", action="store_true", default=False, help="Use aria2c utility if possible. (EXPERIMENTAL)")
        command_parser.add_argument("--trace_appid", "--trace-appid", new_option_string="--trace-appid", dest="trace_appid", action=StoreAndDeprecateWarningAction, default=os.environ.get("RUCIO_TRACE_APPID", None), help=argparse.SUPPRESS)
        command_parser.add_argument("--trace_dataset", "--trace-dataset", new_option_string="--trace-dataset", dest="trace_dataset", action=StoreAndDeprecateWarningAction, default=os.environ.get("RUCIO_TRACE_DATASET", None), help=argparse.SUPPRESS)
        command_parser.add_argument(
            "--trace_datasetscope", "--trace-datasetscope", new_option_string="--trace-datasetscope", dest="trace_datasetscope", action=StoreAndDeprecateWarningAction, default=os.environ.get("RUCIO_TRACE_DATASETSCOPE", None), help=argparse.SUPPRESS
        )  # NOQA: E501
        command_parser.add_argument("--trace_eventtype", "--trace-eventtype", new_option_string="--trace-eventtype", dest="trace_eventtype", action=StoreAndDeprecateWarningAction, default=os.environ.get("RUCIO_TRACE_EVENTTYPE", None), help=argparse.SUPPRESS)  # NOQA: E501
        command_parser.add_argument("--trace_pq", "--trace-pq", new_option_string="--trace-pq", dest="trace_pq", action=StoreAndDeprecateWarningAction, default=os.environ.get("RUCIO_TRACE_PQ", None), help=argparse.SUPPRESS)
        command_parser.add_argument("--trace_taskid", "--trace-taskid", new_option_string="--trace-taskid", dest="trace_taskid", action=StoreAndDeprecateWarningAction, default=os.environ.get("RUCIO_TRACE_TASKID", None), help=argparse.SUPPRESS)
        command_parser.add_argument("--trace_usrdn", "--trace-usrdn", new_option_string="--trace-usrdn", dest="trace_usrdn", action=StoreAndDeprecateWarningAction, default=os.environ.get("RUCIO_TRACE_USRDN", None), help=argparse.SUPPRESS)
        command_parser.add_argument("--filter", dest="filter", action="store", help="Filter files by key-value pairs like guid=2e2232aafac8324db452070304f8d745")
        command_parser.add_argument("--scope", dest="scope", action="store", help="Scope if you are using the filter option and no full DID")
        command_parser.add_argument("--metalink", dest="metalink_file", action="store", help="Path to a metalink file")
        command_parser.add_argument("--deactivate-file-download-exceptions", dest="deactivate_file_download_exceptions", action="store_true", help="Does not raise NoFilesDownloaded, NotAllFilesDownloaded or incorrect number of output queue files Exception")  # NOQA: E501
        command_parser.add_argument("--replica-selection", dest="sort", help="Select the best replica using a replica sorting algorithm provided by replica sorter (e.g., random, geoip)")

    def usage_example(self):
        return [
            "$ rucio download --did my/scope:my/file/name --dir .  # Download to your current dir"
        ]

    def _operations(self) -> dict[str, "OperationDict"]:
        return {}

    def __call__(self):
        download(self.args, self.client, self.logger, self.console, self.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import add_lifetime_exception
from rucio.client.commands.command_base import CommandBase

if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace
    from logging import Logger

    from rich.console import Console
    from rich.status import Status

    from rucio.client.client import Client
    from rucio.client.commands.utils import OperationDict


class LifetimeException(CommandBase):
    def __init__(self, client: "Client", args: "Namespace", logger: "Logger", console: "Console", spinner: "Status") -> None:
        super().__init__(client, args, logger, console, spinner)
        self.PARSER_NAME = "lifetime-exception"

    def module_help(self) -> str:
        return "Manage Lifetime Exceptions (to make protections against deletion from reaper daemons)"

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "add": {"call": self.add, "namespace": self.namespace},
        }

    def usage_example(self) -> list[str]:
        return ["$ rucio lifetime-exception add --inputfile myfile.txt --reason 'Needed for my analysis' --expiration 2015-10-30  # Add exceptions for all DIDs listed in myfile.txt"]

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-f", "--inputfile", help="File where the list of datasets requested to be extended are located")
        parser.add_argument("--reason", help="The reason for the extension")
        parser.add_argument("-x", "--expiration", help="The expiration date format YYYY-MM-DD")

    def add(self):
        add_lifetime_exception(self.args, self.client, self.logger, self.console, self.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from rucio.client.commands.account import Account  # noqa: F401
from rucio.client.commands.command_base import CommandBase  # noqa: F401
from rucio.client.commands.config import Config  # noqa: F401
from rucio.client.commands.did import DID  # noqa: F401
from rucio.client.commands.download import Download  # noqa: F401
from rucio.client.commands.lifetime_exception import LifetimeException  # noqa: F401
from rucio.client.commands.replica import Replica  # noqa: F401
from rucio.client.commands.rse import RSE  # noqa: F401
from rucio.client.commands.rule import Rule  # noqa: F401
from rucio.client.commands.scope import Scope  # noqa: F401
from rucio.client.commands.subscription import Subscription  # noqa: F401
from rucio.client.commands.upload import Upload  # noqa: F401



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import add_container, add_dataset, attach, close, delete_metadata, detach, erase, get_metadata, list_content, list_content_history, list_dids, list_parent_dids, reopen, set_metadata, stat, touch
from rucio.client.commands.command_base import CommandBase

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class DID(CommandBase):
    def module_help(self) -> str:
        return "Manage Data IDentifiers. Modify and access specific files and groups of files. DIDs are accessed by the pattern `scope`:`name`, where name can be a wildcard, but scope must be specified"

    def list_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--recursive", dest="recursive", action="store_true", help="List data identifiers recursively")
        parser.add_argument("--filter", help="Filter arguments in form `key=value,another_key=next_value`. Valid keys are name, type")
        parser.add_argument("--short", action="store_true", help="Just dump the list of DIDs")
        parser.add_argument("-d", "--did", nargs=1, help="Data IDentifier pattern")

    def add_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--type", dest='dtype', choices=("container", "dataset"), help="Add collection type DID")
        parser.add_argument("--monotonic", action="store_true", help="Monotonic status to True")
        parser.add_argument("-d", "--did", action="store", help="The name of the dataset to add")
        parser.add_argument("--lifetime", dest="lifetime", action="store", type=int, help="Lifetime in seconds")

    def update_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-d", "--did", dest="dids", nargs="+", help="List of space separated data identifiers")
        parser.add_argument("--rse", "--rse-name", dest='rse', help="Touch Argument: The RSE of the DIDs that are touched")

        parser.add_argument("--touch", action="store_true", help="Update the last updated time to the current time. Requires a RSE to be set")
        parser.add_argument("--close", action="store_true", help="Set a collection-type DID to 'closed', so it cannot have more child DIDs added to it")
        parser.add_argument("--open", action="store_true", help="Set a collection-type DID to 'open', so more DIDs may be added to it as children")

    def show_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-d", "--did", dest="dids", nargs="+", help="List of space separated data identifiers")
        parser.add_argument("--parent", action="store_true", help="List the parents of the DID")

        # Both non-functional, but list_parents complains if not present
        # Planned to re-implement in a future release
        parser.add_argument("--pfn", dest="pfns", nargs="+", help=argparse.SUPPRESS)
        parser.add_argument("--guid", dest="guids", nargs="+", help=argparse.SUPPRESS)

    def remove_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--undo", action="store_true", help="Undo erase DIDs. Only works if has been less than 24 hours since erase operation")
        parser.add_argument("-d", "--did", dest="dids", nargs="+", help="List of space separated data identifiers")

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "List the Data IDentifiers matching certain pattern. Only collection type DIDs are returned by default, use --filter 'type=all' to return all", "namespace": self.list_namespace},
            "show": {"call": self.show, "docs": "List attributes and statuses about data identifiers", "namespace": self.show_namespace},
            "add": {"call": self.add, "docs": "Create a new collection type data identifier", "namespace": self.add_namespace},
            "remove": {"call": self.remove, "docs": "Delete a DID. Can be recovered for up to 24 hours after deletion", "namespace": self.remove_namespace},
            "update": {"call": self.update, "docs": "Touch one or more DIDs and set the last accessed date to the current date", "namespace": self.update_namespace}
        }

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {"content": Content, "metadata": Metadata}

    def usage_example(self) -> list[str]:
        return [
            "$ rucio did list --did user.jdoe:*  # Show all collection level DIDs with the scope user.jdoe",
            "$ rucio did list --short --filter type=CONTAINER --did user.jdoe:* # Show the names of all container type DIDs",
            "$ rucio did list --filter type=all --did user.jdoe:*  # Show all DIDs with the scope user.jdoe",
            "$ rucio did add --type container --did user.jdoe:container_12345  # Create a new container-type did.",
            "$ rucio did remove --did user.jdoe:file_12345  # Disable file_12345. Will be deleted 24 after deletions.",
            "$ rucio did update --touch --did user.jdoe:file_12345  # Update the time the DID has been modified",
            "$ rucio did show --did user.jdoe:file_12345  # Get the stats for file_12345 - account holder, size, expiration, open status, type, etc",
            "$ rucio did show --parent --did user.jdoe:file_12345 # Show all the parent DIDs for file_12345",
        ]

    def list_(self):
        list_dids(self.args, self.client, self.logger, self.console, self.spinner)

    def show(self):
        if self.args.parent:
            self.args.did = self.args.dids[0]
            list_parent_dids(self.args, self.client, self.logger, self.console, self.spinner)
        else:
            stat(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        erase(self.args, self.client, self.logger, self.console, self.spinner)

    def update(self):
        if self.args.touch:
            touch(self.args, self.client, self.logger, self.console, self.spinner)
        elif self.args.open:
            reopen(self.args, self.client, self.logger, self.console, self.spinner)
        elif self.args.close:
            close(self.args, self.client, self.logger, self.console, self.spinner)
        else:
            raise NotImplementedError("No update option specified, please use `rucio did update -h` to see possible update fields")

    def add(self):
        operations = {
            "container": add_container,
            "dataset": add_dataset
        }
        try:
            operations[self.args.dtype](self.args, self.client, self.logger, self.console, self.spinner)
        except KeyError:
            raise NotImplementedError("Can only add collection type DIDs.")


class Content(DID):
    def module_help(self) -> str:
        return "View the content of collection-type DIDs (datasets and containers), and update their open/closed status"

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "Show the contents of a collection-type DI.", "namespace": self.namespace},
            "history": {"call": self.history, "docs": "Show the content history of a collection-type DID, when DIDs were created, modified, or deleted", "namespace": self.namespace},
            "add": {"call": self.add, "docs": "Attach a list of Data IDentifiers (file, dataset or container) to an other Data IDentifier (dataset or container)", "namespace": self.add_namespace},
            "remove": {"call": self.remove, "docs": "Detach a list of Data Identifiers (file, dataset or container) from an other Data Identifier (dataset or container)", "namespace": self.remove_namespace},
        }

    def usage_example(self) -> list[str]:
        return [
            "$ rucio did content list --did user.jdoe:test12345  # Show the content of a collection-like DID",
            "$ rucio did content history --did user.jdoe:test12345  # Show the history of a DID's content",
            "$ rucio did content add --did user.jdoe:file_12345 --to user.jdoe:dataset_123  # Make dataset_123 the parent of file_12345",
            "$ rucio did content remove --did user.jdoe:file_12345 --from user.jdoe:datset_123  # Orphan file_12345"
        ]

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--did", dest="dids", nargs="+", action="store", help="DIDs to manage the contents of, space separated list")
        parser.add_argument("--short", dest="short", action="store_true", help="Only show the list of DIDs")

    def add_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--to", dest="todid", help="Destination Data IDentifier (either dataset or container)")
        parser.add_argument("-f", "--from-file", dest="fromfile", action="store_true", help="Attach the DIDs contained in a file. The file should contain one did per line")
        parser.add_argument("-d", "--did", dest="dids", nargs="+", help="List of space separated data identifiers, or path to file of DIDs when using from-file")

    def remove_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--from", dest="fromdid", help="Target Data IDentifier (dataset or container), from which to detach")
        parser.add_argument("-d", "--did", dest="dids", nargs="+", help="List of space separated data identifiers")

    def list_(self):
        list_content(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        attach(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        detach(self.args, self.client, self.logger, self.console, self.spinner)

    def history(self):
        list_content_history(self.args, self.client, self.logger, self.console, self.spinner)


class Metadata(DID):
    def module_help(self) -> str:
        return "Manage metadata attached to DIDs via a metadata plugin."

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "Show current metadata for a DID", "namespace": self.list_namespace},
            "add": {"call": self.add, "docs": "Add new metadata for a DID"},
            "remove": {"call": self.remove, "docs": "Delete an existing metadata field for a DID"},
        }

    def usage_example(self) -> list[str]:
        return [
            "$ rucio did metadata add --did user.jdoe:test12345 --key project --value MyShinyNewProject # Update the project field for the DID",
            "$ rucio did metadata list --did user.jdoe:test1245  # Show all the metadata for a DID",
            "$ rucio did metadata list --did user.jdoe:test1245 user.jdoe:test67890  # Show all the metadata for both test12345 and test67890",
        ]

    def list_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-d", "--did", nargs="+", dest="dids", help="List of space separated data identifiers")
        parser.add_argument("--plugin", help="Filter down to metadata from specific metadata plugin")

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-d", "--did", help="Single DID to modify")
        parser.add_argument("--plugin", help="Filter down to metadata from specific metadata plugin")
        parser.add_argument("--key", help="Attribute key")
        parser.add_argument("--value", help="Attribute value")

    def list_(self):
        get_metadata(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        set_metadata(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_metadata(self.args, self.client, self.logger, self.console, self.spinner)




'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '36.0.0.post2',
    'branch_nick': 'release-36',
    'revision_id': 'b07377d21b4df2a3ffd4908a32c504b453de3fa8',
    'revno': 13375
}



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio_admin import delete_config_option, get_config, set_config_option
from rucio.client.commands.command_base import CommandBase

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class Config(CommandBase):
    def module_help(self) -> str:
        return "Manage the global settings.\nCAUTION: changing global configurations can have unintended consequences!"

    def namespace(self, parser: "ArgumentParser") -> "ArgumentParser":
        parser.add_argument("-s", "--section", help="Section name")
        parser.add_argument("-o", "--option", help="Option name")
        parser.add_argument("-v", "--value", help="String-encoded value")
        return parser

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "Show the existing config", "namespace": self.namespace},
            "add": {"call": self.add, "docs": "Update the existing configuration settings. WARNING: Changes the global config. Can overwrite existing settings!", "namespace": self.namespace},
            "remove": {"call": self.remove, "docs": "Remove a setting from the configuration. WARNING: Changes the global config", "namespace": self.namespace},
        }

    def usage_example(self) -> list[str]:
        return [
            "$ rucio config add --section limitsscratchdisk --option testlimit --value 30  # Change the existing limitstractdisk section",
            "$ rucio config list --section foo # Show the settings in section foo",
            "$ rucio config list # Show all the different sections of the config",
            "$ rucio config remove --section testsection --option test  # Remove the value in testsection/test",
        ]

    def list_(self):
        get_config(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        set_config_option(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_config_option(self.args, self.client, self.logger, self.console, self.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from argparse import SUPPRESS
from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import list_dataset_replicas, list_file_replicas, list_suspicious_replicas
from rucio.client.commands.bin_legacy.rucio_admin import declare_bad_file_replicas, declare_temporary_unavailable_replicas, quarantine_replicas, set_tombstone
from rucio.client.commands.command_base import CommandBase
from rucio.common.utils import StoreAndDeprecateWarningAction

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class Replica(CommandBase):
    def module_help(self) -> str:
        return "Interact with a Data IDentifier at a specific Rucio Service Element"

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "List replicas for a given DID", "namespace": self.list_namespace},
            "remove": {
                "call": self.remove,
                "docs": "Set a replica for removal by adding a tombstone which will mark the replica as ready for deletion by a reaper daemon",
                "namespace": self.remove_namespace},
        }

    def list_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument(dest='dtype', choices=("file", "dataset"), help="List either the replicas of a file or a dataset (and its contents)")
        parser.add_argument("-d", "--did", dest="dids", nargs="+", action="store", help="List of space separated data identifiers.")
        parser.add_argument("--protocols", help="Protocol used to access a replicas (i.e. https, root, srm)", required=False)
        parser.add_argument(
            "--all-states",
            action="store_true",
            help="To select all replicas (including unavailable ones).\
                Also gets information about the current state of a DID in each RSE",
            required=False,
        )
        parser.add_argument("--pfns", action="store_true", help="Show only the PFNs", required=False)
        parser.add_argument("--domain", help="Force the networking domain. Available options: wan, lan, all", required=False)
        parser.add_argument(
            "--link",
            help="Symlink PFNs with directory substitution.\
                For example: rucio list-file-replicas --rse RSE_TEST --link /eos/:/eos/ scope:datasetname",
            required=False,
        )
        parser.add_argument("--missing", action="store_true", help="To list missing replicas at a RSE-Expression. Must be used with --rses option", required=False)
        parser.add_argument("--metalink", action="store_true", help="Output available replicas as metalink", required=False)
        parser.add_argument("--no-resolve-archives", action="store_true", help="Do not resolve archives which may contain the files", required=False)
        parser.add_argument("--sort", help="Replica sort algorithm. Available options: geoip (default), random", required=False)
        parser.add_argument("--rses", "--rse-exp", dest="rses", help="The RSE filter expression")
        parser.add_argument("--human", default=True, help=SUPPRESS)

        # Dataset options.
        parser.add_argument("--deep", action="store_true", help="Dataset option only: Make a deep check, checking the contents of datasets in datasets")
        parser.add_argument(
            "--csv",
            action="store_true",
            help="Dataset option only: Write output to comma separated values",
        )

    def remove_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-d", "--did", dest="dids", help="DIDs to access, as comma separated values")
        parser.add_argument("--rse", "--rse-name", dest="rse", help="RSE Name")

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {"state": State}

    def usage_example(self) -> list[str]:
        return [
            "$ rucio replica list file --did user.jdoe:test_file  # Show all replicas for user.jdoe:test_file, with their pfn and rse",
            "$ rucio replica list dataset --did user.jdoe:test_dataset  # Show all replicas for the dataset user.jdoe:test_dataset",
            "$ rucio replica list file --did user.jdoe:test_file --pfns  # Show the PFNs for replicas of user.jdoe:test_file",
            "$ rucio replica remove --did user.jdoe:test_file --rse MyRSE  # Mark a replica for deletion by a reaper daemon"
        ]

    def list_(self):
        if self.args.dtype == "file":
            list_file_replicas(self.args, self.client, self.logger, self.console, self.spinner)

        elif self.args.dtype == "dataset":
            list_dataset_replicas(self.args, self.client, self.logger, self.console, self.spinner)
        else:
            raise NotImplementedError("Only 'file' and 'dataset' are listable")

    def remove(self):
        set_tombstone(self.args, self.client, self.logger, self.console, self.spinner)


class State(Replica):
    def module_help(self) -> str:
        return "Manage replica state."

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "List replicas filtered by state (implemented: (suspicious))", "namespace": self.list_namespace},
            "update": {"call": self.update, "docs": "Update the state of a replica. (implemented: (list as (bad, quarantined, or temporarily unavailable)))", "namespace": self.update_namespace},
        }

    def list_namespace(self, parser: "ArgumentParser"):
        parser.add_argument("state_type", choices=("suspicious",), help="State to filter by when listing")
        parser.add_argument("--rses", "--rse-exp", dest="rse_expression", action="store", help="RSE name or expression")  # TODO remap rse_expression to rses (for consistency)
        parser.add_argument("--younger-than", help='List files that have been marked suspicious since the date "younger_than", e.g. 2021-11-29T00:00:00')  # NOQA: E501
        parser.add_argument("--nattempts", dest="nattempts", action="store", help="Minimum number of failed attempts to access a suspicious file")

    def update_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("state_type", choices=("bad", "unavailable", "quarantine"))

        parser.add_argument("--files", nargs="*", dest="files", help="List of items to update. Each can be a PFN (for one replica) or an LFN (for all replicas of the LFN) or a collection DID (for all file replicas in the DID)")
        parser.add_argument("--input-file", dest="inputfile", nargs="?", action="store", help="File containing list of replicas to update state")
        parser.add_argument('--rse', "--rse_name", dest='rse', nargs='?', action='store', help='RSE Name')

        parser.add_argument("--reason", dest="reason", action="store", help="Supply a reason for changing the replica state")

        bad_parser = parser.add_argument_group("Bad")
        bad_parser.add_argument("--lfns", dest="lfns", nargs="?", action="store", help="File containing list of LFNs for bad replicas. Requires --rse and --scope")
        bad_parser.add_argument("--scope", dest="scope", nargs="?", action="store", help="Common scope for bad replicas specified with LFN list, ignored without --lfns")

        unavailable_parser = parser.add_argument_group("Unavailable")
        unavailable_parser.add_argument("--expiration-date", "--duration", new_option_string="--duration", dest="duration", action=StoreAndDeprecateWarningAction, type=int, help="Timeout in seconds when the replicas will become available again")  # NOQA: E501

        parser.add_argument_group("Quarantine")

    def usage_example(self) -> list[str]:
        return [
            "$ rucio replica state update bad --files mock:test --rse RSE  # Declare the replica of did mock:test at RSE to be bad",
            "$ rucio replica state update unavailable --files mock:test --rse RSE --duration 10 # Declare mock:test at RSE to be unavailable for 10 seconds",
            "$ rucio replica state update quarantine --files mock:test --rse RSE  # Apply a quarantine to mock:test",
            "$ rucio replica state list suspicious --rses RSE  # Show all the suspicious replicas at RSE",
        ]

    def list_(self):
        list_suspicious_replicas(self.args, self.client, self.logger, self.console, self.spinner)

    def update(self):
        if self.args.state_type == "bad":
            self.args.listbadfiles = self.args.files
            declare_bad_file_replicas(self.args, self.client, self.logger, self.console, self.spinner)
        elif self.args.state_type == "unavailable":
            self.args.listbadfiles = self.args.files
            if not hasattr(self.args, "duration"):
                raise ValueError("Missing argument '--expiration-date/--duration")

            declare_temporary_unavailable_replicas(self.args, self.client, self.logger, self.console, self.spinner)
        elif self.args.state_type == "quarantine":
            self.args.paths_list = self.args.files
            quarantine_replicas(self.args, self.client, self.logger, self.console, self.spinner)

        else:
            raise NotImplementedError



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import upload
from rucio.client.commands.command_base import CommandBase
from rucio.common.config import config_get_float

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class Upload(CommandBase):
    def module_help(self) -> str:
        return "Upload DIDs"

    def parser(self, parser: "argparse._SubParsersAction[ArgumentParser]") -> None:

        command_parser = parser.add_parser(self.PARSER_NAME, description=self._help(), formatter_class=argparse.RawDescriptionHelpFormatter)

        command_parser.add_argument("--files", nargs="+", dest="args", help="Files and datasets to upload")
        command_parser.add_argument('--rse', "--rse-name", dest='rse', action='store', help='Rucio Storage Element (RSE) name.',)
        command_parser.add_argument("--lifetime", type=int, help="Lifetime of the rule in second.")
        command_parser.add_argument("--expiration-date", help="The date when the rule expires in UTC, format: <year>-<month>-<day>-<hour>:<minute>:<second>. E.g. 2022-10-20-20:00:00")
        command_parser.add_argument("--scope", help="Scope name to assign new files")
        command_parser.add_argument("--impl", help="Transfer protocol implementation to use (e.g: xrootd, gfal.NoRename, webdav, ssh.Rsync, rclone)")
        # The --no-register option is hidden. This is pilot ONLY. Users should not use this. Will lead to unregistered data on storage!
        command_parser.add_argument("--no-register", action="store_true", default=False, help=argparse.SUPPRESS)
        command_parser.add_argument("--register-after-upload", action="store_true", default=False, help="Register the file only after successful upload")
        command_parser.add_argument("--summary", action="store_true", default=False, help="Create rucio_upload.json summary file")
        command_parser.add_argument("--guid", help="Manually specify the GUID for the file")
        command_parser.add_argument("--protocol", help="Force the protocol to use")
        command_parser.add_argument("--pfn", help="Specify the exact PFN for the upload")
        command_parser.add_argument("--name", help="Specify the exact LFN for the upload")
        command_parser.add_argument("--transfer-timeout", type=float, default=config_get_float("upload", "transfer_timeout", False, 360), help="Transfer timeout (in seconds)")
        command_parser.add_argument("--recursive", action="store_true", default=False, help="Convert recursively the folder structure into collections")

    def usage_example(self):
        return [
            "$ rucio upload --files [files1 file2 file3]  --rse MyRSE  # Upload files to a specific RSE",
            "$ rucio upload --files [files1 file2 file3]  --rse MyRSE  --register-after-upload # Upload files to a specific RSE, registering it automatically"
        ]

    def _operations(self) -> dict[str, "OperationDict"]:
        return {}

    def __call__(self):
        upload(self.args, self.client, self.logger, self.console, self.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from argparse import SUPPRESS
from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import list_account_usage
from rucio.client.commands.bin_legacy.rucio_admin import (
    add_account,
    add_account_attribute,
    ban_account,
    delete_account,
    delete_account_attribute,
    delete_limits,
    identity_add,
    identity_delete,
    info_account,
    list_account_attributes,
    list_accounts,
    list_identities,
    set_limits,
    unban_account,
    update_account,
)
from rucio.client.commands.command_base import CommandBase

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class Account(CommandBase):
    def module_help(self) -> str:
        return "Methods to add or change accounts for users, groups, and services. Used to assign privileges."

    def usage_example(self) -> list[str]:
        return [
            "$ rucio account list # List all accounts on the instance",
            "$ rucio account add --account user_jdoe --type USER  # Create a new user account",
            "$ rucio account update --account user_jdoe --email jdoe@cern.ch  # Update jdoe's email",
            "$ rucio account update --account jdoe --ban True  # Ban jdoe",
            "$ rucio account history --account root  # Show all the usage history for the account root",
        ]

    def list_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--type", dest="account_type", help="Account Type", choices={"USER", "GROUP", "SERVICE"})
        parser.add_argument("-a", "--account", dest="account", help="Account name")
        parser.add_argument("--id", dest="identity", action="store", help="Identity (e.g. DN)")
        parser.add_argument("--filters", dest="filters", action="store", help="Filter arguments in form `key=value,another_key=next_value`")

    def add_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--type", dest="accounttype", help="Account Type (USER, GROUP, SERVICE)", required=True)
        parser.add_argument("-a", "--account", dest="account", help="Account name", required=True)
        parser.add_argument("--email", dest="accountemail", help="Add an email address associated with the account")

    def show_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-a", "--account", dest="account", help="Account name", required=True)

    def update_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-a", "--account", help="Account name", required=True)
        parser.add_argument("--email", help="Account email")
        parser.add_argument("--ban", type=bool, choices=(True, False), help='Ban the account, to disable it. Use --ban False to unban.', default=None)

    def remove_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-a", "--account", dest="acnt", action="store", help="Account name", required=True)

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "List accounts", "namespace": self.list_namespace},
            "add": {"call": self.add, "docs": "Add a new account", "namespace": self.add_namespace},
            "show": {"call": self.show, "docs": "Get all stats on an account, including status, account type, and dates of creation and updates", "namespace": self.show_namespace},
            "remove": {"call": self.remove, "docs": "Delete an account", "namespace": self.remove_namespace},
            "update": {"call": self.update, "docs": "Change the basic account settings", "namespace": self.update_namespace},
        }

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {"attribute": Attribute, "limit": Limit, "identity": Identity}

    def list_(self):
        list_accounts(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        add_account(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_account(self.args, self.client, self.logger, self.console, self.spinner)

    def show(self):
        info_account(self.args, self.client, self.logger, self.console, self.spinner)

    def update(self):
        if self.args.ban is not None:
            if self.args.ban:
                ban_account(self.args, self.client, self.logger, self.console, self.spinner)
            else:
                unban_account(self.args, self.client, self.logger, self.console, self.spinner)
        else:
            # Temp fix to allow specific kwargs
            self.args.key = "email"
            self.args.value = self.args.email
            update_account(self.args, self.client, self.logger, self.console, self.spinner)


class Attribute(Account):
    def module_help(self) -> str:
        return "Add additional key/value pairs associated with an account."

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def namespace(self, subparser):
        subparser.add_argument("-a", "--account", help="Account name")
        subparser.add_argument("--key", dest="key", action="store", help="Attribute key")
        subparser.add_argument("--value", dest="value", action="store", help="Attribute value")

    def usage_example(self) -> list[str]:
        return ["$ rucio account attribute list --account jdoe  # Show all attributes for jdoe",
                "$ rucio -v account attribute add --account jdoe --key test_key --value true  # Set test_key = true for jdoe"]

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "List all account attributes"},
            "add": {"call": self.add, "docs": "Add a new attribute to an account or update an existing one"},
            "remove": {"call": self.remove, "docs": "Remove an existing account attribute"},
        }

    def list_(self):
        list_account_attributes(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        add_account_attribute(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_account_attribute(self.args, self.client, self.logger, self.console, self.spinner)


class Limit(Account):
    def module_help(self) -> str:
        return "Manage storage limits for an account at a given RSE."

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-a", "--account", dest="account", help="Account name", required=True)
        parser.add_argument("--rses", "--rse-exp", dest='rse', action="store", help="RSE expression")
        parser.add_argument("--bytes", action="store", help='Value can be specified in bytes ("10000"), with a storage unit ("10GB"), or "infinity"')
        parser.add_argument("--locality", nargs="?", default="local", choices=["local", "global"], help="Global or local limit scope")
        parser.add_argument("--human", default=True, help=SUPPRESS)

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "Show limits and current utilization for an account at a given RSE", "namespace": self.namespace},
            "add": {"call": self.add, "docs": "Add or update limits for an account", "namespace": self.namespace},  # TODO Add and update should become different operations
            "remove": {"call": self.remove, "docs": "Remove all limits for given account/rse/locality", "namespace": self.namespace},
        }

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def usage_example(self) -> list[str]:
        return super().usage_example()

    def list_(self):
        self.args.usage_account = self.args.account
        list_account_usage(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        set_limits(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        delete_limits(self.args, self.client, self.logger, self.console, self.spinner)


class Identity(Account):
    def module_help(self) -> str:
        return "Manage identities on an account."

    def namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("--account", dest="account", action="store", help="Account name", required=True)
        parser.add_argument("--type", dest="authtype", action="store", choices=["X509", "GSS", "USERPASS", "SSH", "SAML", "OIDC"], help="Authentication type")
        parser.add_argument("--id", dest="identity", action="store", help="Identity as a DNs for X509 IDs.")
        parser.add_argument("--email", dest="email", action="store", help="Email address associated with the identity")
        parser.add_argument("--password", dest="password", action="store", help="Password if authtype is USERPASS")
        parser.add_argument("--human", default=True, help=SUPPRESS)

    def _operations(self) -> dict[str, "OperationDict"]:
        return {
            "list": {"call": self.list_, "docs": "Display existing identities for an account", "namespace": self.namespace},
            "add": {"call": self.add, "docs": "Grant identity access to an account", "namespace": self.namespace},
            "remove": {"call": self.remove, "docs": "Revoke identity access for an account", "namespace": self.namespace},
        }

    def usage_example(self) -> list[str]:
        return [
            "$ rucio account identity list --account jdoe  # List all auth identities for jdoe",
            "$ rucio account identity add --account jdoe --type GSS --email jdoe@cern.ch --id jdoe@fnal.ch  # Add a new GSS auth",
            "$ rucio account identity add --account jdoe --type X509 --id 'CN=Joe Doe,CN=707658,CN=jdoe,OU=Users,OU=Organic Units,DC=cern,DC=ch' --email jdoe@cern.ch  # Add a new X509 auth",
        ]

    def implemented_subcommands(self) -> dict[str, type[CommandBase]]:
        return {}

    def list_(self):
        list_identities(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        identity_add(self.args, self.client, self.logger, self.console, self.spinner)

    def remove(self):
        identity_delete(self.args, self.client, self.logger, self.console, self.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from rucio.client.commands.bin_legacy.rucio import list_scopes
from rucio.client.commands.bin_legacy.rucio_admin import add_scope
from rucio.client.commands.command_base import CommandBase

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from rucio.client.commands.utils import OperationDict


class Scope(CommandBase):
    def module_help(self) -> str:
        return "Manage scopes - A namespace partition generally used to separate common data from user data"

    def _operations(self) -> dict[str, "OperationDict"]:
        return {"list": {"call": self.list_, "docs": "Show existing scopes", "namespace": self.list_namespace}, "add": {"call": self.add, "docs": "Create a new scope", "namespace": self.add_namespace}}

    def usage_example(self) -> list[str]:
        return [
            "$ rucio scope add --scope user.jdoe --account jdoe  # Add a new scope, 'user.jdoe' for use with the account jdoe",
            "$ rucio scope list --account jdoe  # List the existing scopes for account jdoe"
        ]

    def list_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-a", "--account", help="Account name for filtering, attribution")

    def add_namespace(self, parser: "ArgumentParser") -> None:
        parser.add_argument("-a", "--account", help="Account name for filtering, attribution", required=True)
        parser.add_argument("-s", "--scope", help="Name of the new scope to add", required=True)

    def list_(self):
        list_scopes(self.args, self.client, self.logger, self.console, self.spinner)

    def add(self):
        add_scope(self.args, self.client, self.logger, self.console, self.spinner)
